Graph

Struct Graph 

pub struct Graph { /* private fields */ }
Structure to represent a nonlinear factor graph

Main usage will be via add_factor to add new factors to the graph. Also of note is the linearize function that returns a linear (aka Gaussian) factor graph.

Since the graph represents a nonlinear least-squares problem, during optimization it will be iteratively linearized about a set of variables and solved iteratively.

   assign_symbols,
   containers::{Graph, FactorBuilder},
   residuals::PriorResidual,
   robust::GemanMcClure,
   traits::*,
   variables::SO2,
};
let mut graph = Graph::new();
graph.add_factor(factor);

Implementations

Source

impl Graph

pub fn new() -> Self

Examples found in repository?
examples/graph_viz.rs (line 25)
24fn main() {
25    let mut graph = Graph::new();
26    let id = SE2::identity();
27
28    graph.add_factor(fac![PriorResidual::new(id.clone()), X(1)]);
29    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(1), X(2))]);
30    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(2), X(3))]);
31
32    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(1), X(4))]);
33    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(2), X(4))]);
34    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(3), X(5))]);
35
36    rerun_viz(&graph);
37}
More examples
Hide additional examples
examples/readme.rs (line 21)
11fn main() {
12    // Make all the values
13    let mut values = Values::new();
14
15    let x = SO2::from_theta(1.0);
16    let y = SO2::from_theta(2.0);
17    values.insert(X(0), SO2::identity());
18    values.insert(X(1), SO2::identity());
19
20    // Make the factors & insert into graph
21    let mut graph = Graph::new();
22    let res = PriorResidual::new(x.clone());
23    let factor = fac![res, X(0)];
24    graph.add_factor(factor);
25
26    let res = BetweenResidual::new(y.minus(&x));
27    let factor = fac![res, (X(0), X(1)), 0.1 as std, Huber::default()];
28    graph.add_factor(factor);
29
30    // Optimize!
31    let mut opt: GaussNewton = GaussNewton::new_default(graph);
32    let result = opt.optimize(values).unwrap();
33    println!("Results {result:#}");
34}
examples/serde.rs (line 31)
7fn main() {
8    // ------------------------- Serialize values ------------------------- //
9    let x = SO2::from_theta(0.6);
10    let y = SE2::new(1.0, 2.0, 0.3);
11    let mut values = Values::new();
12    values.insert(X(0), x.clone());
13    values.insert(Y(1), y.clone());
14
15    println!("------ Serializing Values ------");
16
17    let serialized = serde_json::to_string_pretty(&values).unwrap();
18    println!("serialized = {serialized}");
19
20    // Convert the JSON string back to a Point.
21    let deserialized: Values = serde_json::from_str(&serialized).unwrap();
22    println!("deserialized = {deserialized:#}");
23
24    // ------------------------- Serialize graph ------------------------- //
25    let prior = PriorResidual::new(x);
26    let bet = BetweenResidual::new(y);
27
28    let prior = fac![prior, X(0), 0.1 as cov, GemanMcClure::default()];
29    let bet = fac![bet, (Y(0), Y(1)), 10.0 as cov];
30
31    let mut graph = Graph::new();
32    graph.add_factor(prior);
33    graph.add_factor(bet);
34
35    println!("\n\n------ Serializing Graph ------");
36
37    let serialized = serde_json::to_string_pretty(&graph).unwrap();
38    println!("serialized = {serialized}");
39
40    let deserialized: Graph = serde_json::from_str(&serialized).unwrap();
41    println!("deserialized = {deserialized:#?}");
42}
examples/gps.rs (line 88)
87fn main() {
88    let mut graph = Graph::new();
89
90    // Add odometry factors
91    let res = BetweenResidual::new(SE2::new(0.0, 2.0, 0.0));
92    let odometry_01 = fac![res.clone(), (X(0), X(1)), (0.1, 0.2) as cov];
93    let odometry_12 = fac![res, (X(1), X(2)), (0.1, 0.2) as cov];
94    graph.add_factor(odometry_01);
95    graph.add_factor(odometry_12);
96
97    // Add gps factors
98    let g0 = fac![GpsResidual::new(0.0, 0.0), X(0), 1.0 as std];
99    let g1 = fac![GpsResidual::new(2.0, 0.0), X(1), 1.0 as std];
100    let g2 = fac![GpsResidual::new(4.0, 0.0), X(2), 1.0 as std];
101    graph.add_factor(g0);
102    graph.add_factor(g1);
103    graph.add_factor(g2);
104
105    // Make values
106    let mut values = Values::new();
107    values.insert(X(0), SE2::new(1.0, 2.0, 3.0));
108    values.insert(X(1), SE2::identity());
109    values.insert(X(2), SE2::identity());
110
111    // // These will all compile-time error
112    // // mismatched symbol-variable types
113    // values.insert(X(5), VectorVar2::identity());
114    // // wrong number of keys
115    // let f = fac![GpsResidual::new(0.0, 0.0), (X(0), X(1))];
116    // // wrong noise-model dimension
117    // let n = factrs::noise::GaussianNoise::<5>::from_scalar_sigma(0.1);
118    // let f = fac![GpsResidual::new(0.0, 0.0), X(0), n];
119    // // mismatched symbol-variable types
120    // assign_symbols!(Y : VectorVar2);
121    // let f = fac![GpsResidual::new(0.0, 0.0), Y(0), 0.1 as std];
122
123    // optimize
124    let mut opt: GaussNewton = GaussNewton::new_default(graph);
125    let result = opt.optimize(values).expect("Optimization failed");
126
127    println!("Final Result: {result:#?}");
128}
pub fn with_capacity(capacity: usize) -> Self

pub fn at(&self, idx: usize) -> &Factor

pub fn add_factor(&mut self, factor: Factor)

Examples found in repository?
examples/graph_viz.rs (line 28)
24fn main() {
25    let mut graph = Graph::new();
26    let id = SE2::identity();
27
28    graph.add_factor(fac![PriorResidual::new(id.clone()), X(1)]);
29    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(1), X(2))]);
30    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(2), X(3))]);
31
32    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(1), X(4))]);
33    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(2), X(4))]);
34    graph.add_factor(fac![BetweenResidual::new(id.clone()), (X(3), X(5))]);
35
36    rerun_viz(&graph);
37}
More examples
Hide additional examples
examples/readme.rs (line 24)
11fn main() {
12    // Make all the values
13    let mut values = Values::new();
14
15    let x = SO2::from_theta(1.0);
16    let y = SO2::from_theta(2.0);
17    values.insert(X(0), SO2::identity());
18    values.insert(X(1), SO2::identity());
19
20    // Make the factors & insert into graph
21    let mut graph = Graph::new();
22    let res = PriorResidual::new(x.clone());
23    let factor = fac![res, X(0)];
24    graph.add_factor(factor);
25
26    let res = BetweenResidual::new(y.minus(&x));
27    let factor = fac![res, (X(0), X(1)), 0.1 as std, Huber::default()];
28    graph.add_factor(factor);
29
30    // Optimize!
31    let mut opt: GaussNewton = GaussNewton::new_default(graph);
32    let result = opt.optimize(values).unwrap();
33    println!("Results {result:#}");
34}
examples/serde.rs (line 32)
7fn main() {
8    // ------------------------- Serialize values ------------------------- //
9    let x = SO2::from_theta(0.6);
10    let y = SE2::new(1.0, 2.0, 0.3);
11    let mut values = Values::new();
12    values.insert(X(0), x.clone());
13    values.insert(Y(1), y.clone());
14
15    println!("------ Serializing Values ------");
16
17    let serialized = serde_json::to_string_pretty(&values).unwrap();
18    println!("serialized = {serialized}");
19
20    // Convert the JSON string back to a Point.
21    let deserialized: Values = serde_json::from_str(&serialized).unwrap();
22    println!("deserialized = {deserialized:#}");
23
24    // ------------------------- Serialize graph ------------------------- //
25    let prior = PriorResidual::new(x);
26    let bet = BetweenResidual::new(y);
27
28    let prior = fac![prior, X(0), 0.1 as cov, GemanMcClure::default()];
29    let bet = fac![bet, (Y(0), Y(1)), 10.0 as cov];
30
31    let mut graph = Graph::new();
32    graph.add_factor(prior);
33    graph.add_factor(bet);
34
35    println!("\n\n------ Serializing Graph ------");
36
37    let serialized = serde_json::to_string_pretty(&graph).unwrap();
38    println!("serialized = {serialized}");
39
40    let deserialized: Graph = serde_json::from_str(&serialized).unwrap();
41    println!("deserialized = {deserialized:#?}");
42}
examples/gps.rs (line 94)
87fn main() {
88    let mut graph = Graph::new();
89
90    // Add odometry factors
91    let res = BetweenResidual::new(SE2::new(0.0, 2.0, 0.0));
92    let odometry_01 = fac![res.clone(), (X(0), X(1)), (0.1, 0.2) as cov];
93    let odometry_12 = fac![res, (X(1), X(2)), (0.1, 0.2) as cov];
94    graph.add_factor(odometry_01);
95    graph.add_factor(odometry_12);
96
97    // Add gps factors
98    let g0 = fac![GpsResidual::new(0.0, 0.0), X(0), 1.0 as std];
99    let g1 = fac![GpsResidual::new(2.0, 0.0), X(1), 1.0 as std];
100    let g2 = fac![GpsResidual::new(4.0, 0.0), X(2), 1.0 as std];
101    graph.add_factor(g0);
102    graph.add_factor(g1);
103    graph.add_factor(g2);
104
105    // Make values
106    let mut values = Values::new();
107    values.insert(X(0), SE2::new(1.0, 2.0, 3.0));
108    values.insert(X(1), SE2::identity());
109    values.insert(X(2), SE2::identity());
110
111    // // These will all compile-time error
112    // // mismatched symbol-variable types
113    // values.insert(X(5), VectorVar2::identity());
114    // // wrong number of keys
115    // let f = fac![GpsResidual::new(0.0, 0.0), (X(0), X(1))];
116    // // wrong noise-model dimension
117    // let n = factrs::noise::GaussianNoise::<5>::from_scalar_sigma(0.1);
118    // let f = fac![GpsResidual::new(0.0, 0.0), X(0), n];
119    // // mismatched symbol-variable types
120    // assign_symbols!(Y : VectorVar2);
121    // let f = fac![GpsResidual::new(0.0, 0.0), Y(0), 0.1 as std];
122
123    // optimize
124    let mut opt: GaussNewton = GaussNewton::new_default(graph);
125    let result = opt.optimize(values).expect("Optimization failed");
126
127    println!("Final Result: {result:#?}");
128}
pub fn len(&self) -> usize

Examples found in repository?
examples/g2o.rs (line 80)
54fn main() -> Result<(), Box<dyn std::error::Error>> {
55    // ---------------------- Parse Arguments & Load data ---------------------- //
56    let mut args: Vec<String> = env::args().collect();
57    match args.len() {
58        2 => {
59            args.push(String::from("gauss"));
60            args.push(String::from("arrows"));
61        }
62        3 => {
63            args.push(String::from("arrows"));
64        }
65        4 => {}
66        _ => {
67            println!(
68                "Usage: {} <g2o file> <optimizer: [gauss|leven|gnc] = gauss> <vis: [points|arrows] = arrows>",
69                args[0]
70            );
71            return Ok(());
72        }
73    }
74
75    pretty_env_logger::init();
76
77    // Load the graph from the g2o file
78    let filename = &args[1];
79    let (graph, init) = load_g20(filename);
80    println!("File loaded, {} factors", graph.len());
81
82    let obj = &args[3];
83    let dim = if init.filter::<SE2>().count() != 0 {
84        "se2"
85    } else if init.filter::<SE3>().count() != 0 {
86        "se3"
87    } else {
88        panic!("Graph doesn't have SE2 or SE3 variables");
89    };
90
91    // Make optimizer
92    let mut optimizer: Box<dyn Optimizer> = match args[2].as_str() {
93        "gauss" => {
94            let params = BaseOptParams {
95                max_iterations: 200,
96                error_tol_relative: 1e-4,
97                ..Default::default()
98            };
99            let mut opt = GaussNewton::new(params, graph);
100            rerun_init(&mut opt, dim, obj);
101            Box::new(opt)
102        }
103        "leven" => {
104            let mut params = LevenParams::default();
105            params.base.max_iterations = 200;
106            params.base.error_tol_relative = 1e-4;
107            params.min_model_fidelity = -1e4;
108            let mut opt = LevenMarquardt::new(params, graph);
109            rerun_init(&mut opt, dim, obj);
110            Box::new(opt)
111        }
112        #[allow(clippy::field_reassign_with_default)]
113        "gnc" => {
114            let mut params: GncParams<LevenMarquardt> = GncParams::default();
115            params.mu_step_size = 1.4;
116            params.base.max_iterations = 200;
117            params.inner.base.max_iterations = 5;
118            params.inner.base.error_tol_absolute = 1e-4;
119            params.inner.base.error_tol_relative = 1e-4;
120            let mut opt: GraduatedNonConvexity<GncGemanMcClure, _> =
121                GraduatedNonConvexity::new(params, graph);
122            rerun_init(&mut opt, dim, obj);
123            Box::new(opt)
124        }
125        _ => {
126            println!("Optimizer not recognized");
127            return Ok(());
128        }
129    };
130
131    // ------------------------- Optimize ------------------------- //
132    let start = Instant::now();
133    let result = optimizer.optimize(init);
134    let duration = start.elapsed();
135
136    match result {
137        Ok(_) => println!("Optimization converged!"),
138        Err(_) => println!("Optimization failed!"),
139    }
140    println!("Optimization took: {duration:?}");
141    Ok(())
142}
pub fn is_empty(&self) -> bool

pub fn error(&self, values: &Values) -> dtype

pub fn linearize(&self, values: &Values) -> LinearGraph

pub fn sparsity_pattern(&self, order: ValuesOrder) -> GraphOrder

pub fn iter(&self) -> Iter<'_, Factor>

pub fn iter_mut(&mut self) -> IterMut<'_, Factor>